
from staq.stack import Stack, StackFrame, StackCell 
from staq.function import Function, FunctionArgument
from staq.helper import *
from termcolor import colored
from cliify import commandParser, command
from cliify.splitHelper import splitWithEscapes

import re
import json
import yaml



sizeDict = {
    "void": 0,
    "char": 1,
    "unsigned char": 1,
    "signed char": 1,
    "short": 2,
    "unsigned short": 2,
    "int": 4,
    "unsigned int": 4,
    "long": 4,
    "unsigned long": 4,
    "long long": 8,
    "unsigned long long": 8,
    "float": 4,
    "double": 8,
    "long double": 16,  # This can vary, but 16 bytes is a common size
    "int8": 1,
    "uint8": 1,
    "int16": 2,
    "uint16": 2,
    "int32": 4,
    "uint32": 4,
    "int64": 8,
    "uint64": 8,
    "va_list": 4,
}


class Register():
    def __init__(self, name, size, description):
        self.name = name
        self.size = size
        self.type = type
        self.value = None
        self.description = None
        self.note = None


@commandParser
class Architecture():
    """
        Base Class for calling conventions
    """
    def __init__(self, name, session = None, stack = None):
        self.name = name
        self.endian = 'little'
        self.registers = {}
        self.session = session
        self.stack : Stack = stack
        self.wordSize = 4

    def clear(self):

        for key in self.registers:
            self.registers[key].value = None


    def setInstructionPointer(self, val):
        pass

    def setFramePointer(self, val):
        pass

    def setReturnValue(self, val):
        pass

    @command(completions={'arg': ['frame']})
    def pop(self, arg: str = None):
        """
            Pop a value from the stack
            
            pop frame
            pop eax
            pop {eax,ebx,ecx}
            pop { r0-r3, r5}
        """
        if arg == 'frame':
            return self.popFrame()
        elif arg in self.registers:
            self.setReg(arg, self.stack.pop())
        elif arg and  arg.startswith('{') and arg.endswith('}'):
            arg = arg.replace('{','[').replace('}',']')
            parsed = yaml.load(arg, Loader=yaml.FullLoader)
            for x in parsed:
                if x in self.registers:
                    self.setReg(x, self.stack.pop())
                elif '-' in x:
                    start_end = x.split('-')
                    if len(start_end) == 2:
                        start = start_end[0].strip()
                        end = start_end[1].strip()
                        if start in self.registers and end in self.registers:
                            startIndex = list(self.registers.keys()).index(start)
                            endIndex = list(self.registers.keys()).index(end)
                            for i in range(startIndex, endIndex + 1):
                                regName = list(self.registers.keys())[i]
                                self.setReg(regName, self.stack.pop())
        elif arg.isnumeric():
            self.stack.pop(int(arg))
        else:
            self.stack.pop()

    @command(completions={'arg': ['frame']})
    def push(self, arg: str = None, label: str = None, note: str = None, size: int = 0):
        """
            Push a value to the stack

            push frame
            push eax
            push r0 
            push {eax,ebx,ecx}
            push { r0-r3, r5}
        """

        cell = StackCell(label=label)

        if arg == 'frame':
            return self.pushFrame()
        elif arg in self.registers:
            cell.setWords(self.getReg(arg))
            cell.words[0].note = note
            self.stack.push(cell)
        elif arg and arg.startswith('{') and arg.endswith('}'):
            arg = arg.replace('{','[').replace('}',']')
            parsed = yaml.load(arg, Loader=yaml.FullLoader)
            for x in parsed:
                cell = StackCell(label=label)
                if x in self.registers:
                    cell.setWords(self.getReg(x))
                    cell.words[0].note = note
                    self.stack.push(cell)
                elif '-' in x:
                    start_end = x.split('-')
                    if len(start_end) == 2:
                        start = start_end[0].strip()
                        end = start_end[1].strip()
                        if start in self.registers and end in self.registers:
                            startIndex = list(self.registers.keys()).index(start)
                            endIndex = list(self.registers.keys()).index(end)
                            for i in range(startIndex, endIndex + 1):
                                cell = StackCell(label=label)
                                regName = list(self.registers.keys())[i]
                                cell.setWords(self.getReg(regName))
                                cell.words[0].note = note
                                self.stack.push(cell)


        else:
            #try to parse the arg as a number
            try:
                cell.setWords(arg)
                if size > 0:
                    cell.setSize(size)
                cell.words[0].note = note
                self.stack.push(cell)
            except ValueError:
                pass

    @command
    def local(self, line: str,note: str = None):
        """
            Add a local variable to the stack using line

            int singleInt = 4
            int array[10] = {0,1,2,3,4,5,6,7,8,9}
            int *ptr
            char *ptr[10]
            float* ptr[10]
        """

        cell = StackCell(note=note)

        #split the line into parts
        left_right = splitWithEscapes(line, '=')

        if len(left_right) == 1:
            left = left_right[0]
            right = ""
        elif len(left_right) == 2:
            left = left_right[0]
            right = left_right[1]

        left = left.strip()
        right = right.strip()
        elSizeBytes = 4
        elCount = 1
        if right != "":
            right = right.replace('{','[').replace('}',']')
            if right.startswith('[') and right.endswith(']'):
                parsed = yaml.load(right)
                if isinstance(parsed, list):
                    elCount = len(parsed)
            
                cell.setWords(parsed)
            else:
                cell.setWords(right)
        
        leftParts = splitWithEscapes(left, ' ')
        if len(leftParts) == 2:
            type = leftParts[0]
            name = leftParts[1]
            isPointer = False
            if name.startswith('*'):
                isPointer = True
                name = name[1:]
                elSizeBytes = 4
            
            if type.endswith('*'):
                isPointer = True
                type = type[:-1]
                elSizeBytes = 4
            
            if not isPointer:
                if type in sizeDict:
                    elSizeBytes = sizeDict[type]
                else:
                    raise ValueError(f"Unknown type: {type}")

            regex = re.compile(r"(\S+)\[(\d+)\]")
            match = regex.match(name)
            if match:
                name = match.group(1)
                declaredSize = int(match.group(2))
                if declaredSize > elCount:
                    elCount = declaredSize
        
        cell.setSize(elSizeBytes * elCount / self.wordSize)
        cell.label = name

        self.stack.push(cell)
                
    @command
    def int(self, line, note: str = None):
        self.local("int " + line, note)
    
    @command
    def char(self, line, note: str = None):
        self.local("char " + line, note)
    @command
    def short(self, line, note: str = None):
        self.local("short " + line, note)
    @command
    def long(self, line, note: str = None):
        self.local("long " + line, note)
    @command
    def float(self, line, note: str = None):
        self.local("float " + line, note)
    @command
    def double(self, line, note: str = None):
        self.local("double " + line, note)


    
    def popFrame(self):
       return self.stack.popFrame()
    
    def pushFrame(self, frame):
        self.stack.pushFrame(frame)

    def getReg(self, key):
        if key in self.registers:
            return self.registers[key].value
        
        return None
    
    def setReg(self,key, value, note = None, create=True):

        if key in self.registers:
            self.registers[key].value = value
            self.registers[key].note = note
        else:
            if create:
                self.registers[key] = Register(key, 4, note)
                self.registers[key].value = value
                self.registers[key].note = note
            else:
                raise ValueError(f"Register {key} not found")

    @command
    def leave(self):

        if self.stack.currentFrame:
            while self.stack.currentFrame.length() > 1:
                self.stack.pop()
            
            while self.stack.currentFrame.cells[0].size > 1:
                self.stack.pop()

    @command
    def ret(self, value = None):
        """
            Return from the current function
        """
        word = self.stack.pop()

        if value:
            self.setReturnValue(value)


        if word and word.value:
            self.jmp( word.value)
        else:
            self.jmp('???')

    def markArgs(self): 

        self.stack.argMarks = {}

        if self.stack.currentFrame and self.stack.currentFrame.function:
            func = self.stack.currentFrame.function

            pointer = self.stack.currentFrame.basePointer
            pointer+= 1 #skip the return address
            for i in range(len(func.args)):
                
                self.stack.argMarks[pointer] = func.args[i].name
                pointer += func.args[i].size


    @command
    def jmp(self, address):

        val = address
        self.setInstructionPointer(val)

        offset = 0

        if val.startswith('<') and val.endswith('>'):
            val = val[1:-1]

        val = val.strip()

        if '+' in val:
            
            val, offset = val.split('+')


        if self.stack.currentFrame and self.stack.currentFrame.function and self.stack.currentFrame.function.name == val:
            pass 
        else:
            lastWord = self.stack.pop()
            func = self.session.functions.get(val, Function(val))
            self.stack.pushFrame(StackFrame(val,func))
            cell = StackCell('ret', [lastWord])

            self.stack.push(cell)

        self.markArgs()

    def updateRegs(self):
        self.setReg('esp', self.stack.pointer)


    def registersToAnsi(self, width = 30, color='yellow'):

        #  ┌───────────────────────────────────┐
        #  │ esp: <value>                      │
        #  │ ebp: <value>                      │
        #  │ eax: <value>                      │
        #  │ ebx: <value>                      │
        #  └───────────────────────────────────┘

        self.updateRegs()

        #get the max length of the register names
        maxNameLength = 0
        for key in self.registers:
            if len(key) > maxNameLength:
                maxNameLength = len(key)
        
        maxValLength =  width - (maxNameLength + 5)

        out = colored(f"┌{'─' * maxNameLength}{'─' * (maxValLength +5)}┐\n", color)
        for key in self.registers:
            if self.registers[key].value:

                if isinstance(self.registers[key].value, int):
                    strVal = hex(self.registers[key].value)
                else:
                    strVal = str(self.registers[key].value)

                truncVal = strVal.ljust(maxValLength)[:maxValLength]
                out += colored(f"│ {key.rjust(maxNameLength)} : {truncVal} │\n", color)

        out += colored(f"└{'─' * maxNameLength}{'─' * (maxValLength + 5)}┘\n", color)

        return out

    @command(completions={'function': lambda self: self.session.functions.keys()})
    def call(self, function):

        callLine = function.strip()

        regex = re.compile(r"(\S.*?)\((.*?)\)")

        match = regex.match(callLine)

        func : Function = None

        if match:

            functionName = match.group(1)

            args = match.group(2)

        else:
            functionName = callLine
            args = ""

        if args.strip() == "":
            args = []
        else:
            argStr = match.group(2)
            args = splitWithEscapes(argStr,',')


        # If the function does not exist, create it and assume all args are ints
        if functionName not in self.session.functions:
            newFunction = Function(functionName)

            for i in range(len(args)):
                newFunction.addArgument(FunctionArgument('int',f'arg{i+1}'))

            self.session.functions[functionName] = newFunction

        func  = self.session.functions[functionName]
        
        for i in reversed(range(len(args))):
            label = f'arg{i+1}'
            if i < len(func.args) and func.args[i].name: 
                label = f"{func.args[i].name}"
            

            cell = StackCell(  label, [args[i].strip()])
            self.stack.push(cell)

        currentFunction = self.session.getCurrentFunction()

        
        retCell = StackCell('ret', ['<???>'])
        if currentFunction:

            if isinstance(currentFunction, Function):
                
                retCell.setWords(f'<{currentFunction.name}+??>')
            else:
                retCell.setWords(f'<{currentFunction}+??>')

        self.stack.push(retCell)

        self.jmp(functionName)

        if func:
            for local in func.locals:
                cell = self.session.tryParseLocalVar(local)
                if cell:
                    self.stack.push(cell)


        

    




def splitWithEscapes(string, delimiter, escape_pairs = ['""',"{}","[]","<>","()"], strip = True):
    """
    Split a string by a delimiter, but ignore delimiters that are inside escape_pairs.
    """
    split = []
    counts = {}

    for pair in escape_pairs:
        counts[pair] = 0

    start = 0
    current = ""

    def processEscapes(char):
        for key in escape_pairs:
            if key[0] == key[1] and char == key[0]:
                #if the escape pair is the same character, then we toggle the escape at each instance
                counts[key] = counts[key] ^ 1
            else:
                if char == key[0]:
                    counts[key] += 1
                elif char == key[1]:
                    counts[key] -= 1
    
    def isEscaped():
        for key in escape_pairs:
            if counts[key] > 0:
                return True
        return False
        
    current = ''
    for i in range(len(string)):
        
        if string[i] == delimiter and not isEscaped() and current != '':

            if strip:
                current = current.strip()
            split.append(current)
            current = ''
        else:
            processEscapes(string[i])
            current += string[i]
    
    if current != '':
        if strip:
            current = current.strip()
        split.append(current)
    
    return split




from staq.archs import X86_cdecl, Arm32
from staq.stack import Stack, StackFrame, StackCell
from staq.function import Function, FunctionArgument, parseCFile
import re
import yaml
import json
import argparse
import re

from importlib.resources import files, as_file
import builtins
from cliify import command, commandParser
from cliify.splitHelper import splitWithEscapes


sizeDict = {
    "void": 0,
    "char": 1,
    "unsigned char": 1,
    "signed char": 1,
    "short": 2,
    "unsigned short": 2,
    "int": 4,
    "unsigned int": 4,
    "long": 4,
    "unsigned long": 4,
    "long long": 8,
    "unsigned long long": 8,
    "float": 4,
    "double": 8,
    "long double": 16,  # This can vary, but 16 bytes is a common size
    "int8": 1,
    "uint8": 1,
    "int16": 2,
    "uint16": 2,
    "int32": 4,
    "uint32": 4,
    "int64": 8,
    "uint64": 8,
    "va_list": 4,
}


def nullPrint(*args, **kwargs):
    pass

builtins.print = nullPrint

@commandParser( subparsers=['arch'],collapsed=True)
class StackSession():
    def __init__(self, arch = None):
        self.stack = Stack()
        self.arch = None
        #self.arch = X86_cdecl(session=self, stack=self.stack)
        self.functions = {}
        self.parser = argparse.ArgumentParser(description='Stack Visualizer', add_help=False)
        self.subparsers = None
        self.loadLibcFuntions()
        self.commands = []
        self.update = False
        self.status = ""
        self.cmdIdx = 0

    def setArch(self, arch):
        if arch == 'x86':
            self.arch = X86_cdecl(session=self, stack=self.stack)
        elif arch == 'arm':
            self.arch = Arm32(session=self, stack=self.stack)
        else:
            self.arch = X86_cdecl(session=self, stack=self.stack)

    def addHistory(self,cmd):

        if self.cmdIdx < len(self.commands):
            self.commands = self.commands[:self.cmdIdx]

        self.commands.append(cmd)
        self.cmdIdx+=1

    def stepForward(self):

        if self.cmdIdx < len(self.commands):
            self.parseCommand(self.commands[self.cmdIdx], addHistory=False)
            self.cmdIdx +=1

    def stepBack(self):

        if self.cmdIdx > 0:

            self.cmdIdx -= 1
            self.stack.clear()
            
            for i in range(0,self.cmdIdx):
                self.parseCommand(self.commands[i], addHistory=False)

    def print(self,text):
        print(text)

    def refreshOutput(self):

        ansi = self.stack.toAnsi( showAddress=True)
        print(ansi)

    def loadYaml(self, obj):


        if isinstance(obj, str):
            with open(obj) as f:
                obj = yaml.safe_load(f)


        order = 'normal'

        if 'stackBase' in obj:
            self.stack.baseAddress = obj['stackBase']
        
        if 'order' in obj:
            order = obj['order']

        if 'functions' in obj:
            for func in obj['functions']:
                self.addFunction(Function.fromString(func))
    

        if 'stack' in obj:

            nodes = []
            if order == 'normal':
                nodes = reversed(obj['stack'])
            else:
                nodes = obj['stack']

            for node in nodes:
                
               if isinstance(node, dict) and 'function' in node:
                   #Create a new frame 
                   frame = StackFrame.fromObj(node, order= order)
                   self.stack.pushFrame(frame)
                   self.stack.currentFrame = None
               else:
                    cell = StackCell.fromObj(node)
                    self.stack.push(cell)
        
        self.stack.applyAddresses()


    
    def getCurrentFunction(self):

        if self.stack.currentFrame:

            if self.stack.currentFrame.function:
                return self.stack.currentFrame.function
            
        
        return None

    def addFunction(self, function):
        self.functions[function.name] = function


    def loadLibcFuntions(self):
        
        data_path = files('staq.data').joinpath('libc.c')
        libc_functions = []
        with as_file(data_path) as f:
            with open(f) as libc_file:
                functions = parseCFile(libc_file.read())
                for func in functions:
                    self.addFunction(func)


   
    @command
    def frame(self, name, color = None):
        """
            frame <name> --color <color>
        """
        frame = StackFrame(name)
        if color:
            frame.color = color
        self.stack.pushFrame(frame)



    def parseFunction(self, line):

        function = Function.fromString(line)
        self.addFunction(function)


    # def functionCmd(self,args):
    #     line = " ".join(args.function)
    #     function = Function.fromString(line)
    #     self.addFunction(function)
    
    @command( completions={'file': ['$file']} )
    def save(self, file = None):

        filename = file




        if filename.endswith('.html'):
            with open(filename, "w") as f:
                f.write(self.stack.toHtml())

        elif filename.endswith('.png'):

            try:
                self.stack.generatePng(filename)
            except Exception as e:
                self.print("Failed to generate image. This is likely due to not having 'chrome' installed which is required for html2image. Please install chrome and try again.")

        else:
            with open(filename, "w") as f:
                for cmd in self.commands:
                    if not cmd.startswith("save"):
                        if not cmd.endswith("\n"):
                            cmd += "\n"
                        f.write(cmd)


                
        

    @command( completions={'file': ['$file']} )
    def load(self, file: str = None):
        filename = file
        self.update = False

        if filename.endswith('.c'):

            with open(filename) as f:
                functions = parseCFile(f.read())

            self.status = f"Loaded {len(functions)} functions from {filename}"

            for func in functions:
                self.addFunction(func)

    def getCellRefs(self) -> list:
        """
        Get a list of all cell references in the stack
        """
        cells = self.stack.getCells()
        refs = []

        for cell in cells:
            refs.append(f"&{cell.label}")

        return refs

    @command( completions={'file': ['$file'], 'data': lambda  self: self.getCellRefs(), 'address': lambda  self: self.getCellRefs(),  } )
    def write(self, data = None, address = None, file = None ):
        """
            write [data1,data2,data3] , address: &cell+4
            write &cell+16: [data1,data2,data3]
        
        """
        

        if file:
            #read file as bytes 
            with open(file, "rb") as f:
                data = f.read()
                data = list(data)


        if isinstance(data, str):
            if not data.startswith("[") and not data.endswith("]"):
                data = f"[{data}]"
            data = yaml.safe_load(data) #use yaml to parse the list so we get the correct types/escapes

        self.stack.write(data, address)

    @command( completions={'address': lambda  self: self.getCellRefs(), 'note': lambda  self: self.getCellRefs(), 'color': ['red', 'green', 'blue', 'yellow', 'purple'] } )
    def note(self, note = None, address = None, color = 'red'):
        """
            note 0x1234: note on local variable
            note 0x1234: note on local variable --color red
        """

        self.stack.setNote(note=note, address=address, color=color)

    @command(completions={'function': lambda self: self.functions.keys()})
    def run(self, function, limit = 20):

        called = True

        functionCall = None
        
        if function:
            function = function.strip()
            if function == "":
                function = None

        if function:
            functionCall = function
        elif self.stack.currentFrame and self.stack.currentFrame.function:
            function = self.stack.currentFrame.function

            if len(function.calls) > 0:
                functionCall  = function.calls[0]

        if not functionCall:
            functionCall = "main()"

        self.arch.call(functionCall)
        limit -= 1

        while  called and (limit > 0):
            
            if self.stack.currentFrame and self.stack.currentFrame.function: 
                function = self.stack.currentFrame.function

                if len(function.calls) > 0:
                    call = function.calls[0]
                    self.arch.call(call)
                    called = True

            limit -= 1


    @command
    def clear(self):
        self.stack.clear()


    def tryParseCommand(self, line, addHistory = True):
        """
        Try to parse a command. If it fails, try to parse with calling convention
        """
        plainCmd = line.split(" ")[0]

        if addHistory and plainCmd not in ['save','s', 'load','l']:
            self.addHistory(line)


        cmds = splitWithEscapes(line,";")

        for cmd in cmds:
            words = cmd.split(" ")
            firstWord = words[0]

            #If first word is a var declaration for a pointer, move the * to the beginning of the name 
            # i.e int* foo = 0x1234 becomes int *foo = 0x1234
            # This is to allow the command parser to work with the declaration
            if firstWord.endswith("*"):
                remaining = cmd[len(firstWord):].lstrip()
                firstWord = firstWord[:-1]
                cmd = firstWord + " *" + remaining
            
            if firstWord in ["write","note"]:
                """
                    if using shorthand for write, i.e. write &cell+16: [data1,data2,data3]
                """

                if (words[1].startswith("&") or words[1].startswith("0x")) and words[1].endswith(":"):
                    #write &cell+16: [data1,data2,data3]
                    cmd = f'{firstWord} {" ".join(words[2:])} , address: {words[1]}'
                    

                

            self.parseCommand(cmd.strip())
            self.refreshOutput()





    def showHelp(self, command):
            
            if command in self.subparsers.choices:
                self.print(self.subparsers.choices[command].format_help() + "\n\n [Enter] to continue...")
            else:
                self.print(self.parser.format_help() + "\n\n [Enter] to continue...")

            #self.print("[Enterr] to continue...")

    def tryParseLocalVar(self, line):
        
        parts = line.split("=")

        
        revar = re.compile(r"(\w[\w\s\*]*\w)\s+(\w+)(\[(\d*)\])?")

        ptr = False

        decl = parts[0]
        if "*" in decl:
            ptr = True
            decl = decl.replace("*","")

        match = revar.match(decl)
        val = []
        if len(parts) > 1:
            parts[1] = parts[1].split("--")[0]
            val = parts[1].replace(";","")
            val = yaml.safe_load(val)

        if match:
            varType = match.group(1).strip()

            varName = match.group(2)

            arraySize = match.group(4) if match.group(4) else None

            cell = StackCell(varName)

            if "_t" in varType:
                varType = varType.replace("_t","")

            if varType not in sizeDict:
                return None

            
            if ptr:
                sizeBytes = 4
            else:
                sizeBytes = sizeDict.get(varType, 4)

            if arraySize:
                sizeBytes = sizeBytes * int(arraySize)
            
            size = int((sizeBytes + 3) / 4)

            cell.setWords(val)
            cell.setSize(size)
            cell.label = varName


            line = line.replace(match.group(0), 'local')


            return cell

        else:
            return None

